! ========================================================================
! Copyright (C) by GemTalk Systems 1991-2017.  All Rights Reserved
! ========================================================================

expectvalue /Class
doit
Object subclass: 'CPreprocessorToken'
  instVarNames: #( file line source
                    type value peek)
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%
expectvalue /Class
doit
CPreprocessorToken comment: 
'This class is internal to the FFI implementation and 
represents a preprocessor token from a C header file.'
%
expectvalue /Class
doit
CPreprocessorToken category: 'FFI'
%
! ------------------- Remove existing behavior from CPreprocessorToken
expectvalue /Metaclass3       
doit
CPreprocessorToken removeAllMethods.
CPreprocessorToken class removeAllMethods.
%
set compile_env: 0
! ------------------- Class methods for CPreprocessorToken
category: 'other'
classmethod: CPreprocessorToken
args: anArray
	| res |
	(res := self basicNew) initializeArgs: anArray .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
concatenation
	| res |
	(res := self basicNew) initializeConcatenation .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
define: aString
	| res |
	(res := self basicNew) initializeDefine: aString .
    ^ res
%
category: 'other'
classmethod: CPreprocessorToken
directive: aString
	| res |
	(res := self basicNew) initializeDirective: aString .
    ^ res
%
category: 'other'
classmethod: CPreprocessorToken
empty
	| res |
	(res := self basicNew) initializeEmpty .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
identifier: aString
	| res |
	(res := self basicNew) initializeIdentifier: aString .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
int16: anInteger
	| res |
	(res := self basicNew ) initializeInt16: anInteger .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
int32: anInteger
	| res |
	(res := self basicNew) initializeInt32: anInteger .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
int64: anInteger
	| res |
	(res := self basicNew) initializeInt64: anInteger .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
nextFrom: aStream filename: aString line: anInteger
	"Note that this might return nil if no more tokens!"

	^self
		nextFrom: aStream
		filename: aString
		line: anInteger
		onOneLogicalLine: false
%
category: 'other'
classmethod: CPreprocessorToken
nextFrom: aStream filename: aString line: anInteger onOneLogicalLine: aBoolean
	"Note that this might return nil if no more tokens!"

	^self basicNew
		initializeStream: aStream
		filename: aString
		line: anInteger
		singleLogicalLine: aBoolean
%
category: 'other'
classmethod: CPreprocessorToken
nextOnCurrentLineFrom: aStream filename: aString line: anInteger
	"Note that this might return nil if no more tokens!"

	^self
		nextFrom: aStream
		filename: aString
		line: anInteger
		onOneLogicalLine: true
%
category: 'other'
classmethod: CPreprocessorToken
number: aNumber
	| res |
	(res := self basicNew) initializeNumber: aNumber .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
punctuator: aString
	| res |
	(res := self basicNew) initializePunctuator: aString .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
string: aString
	| res |
	(res := self basicNew) initializeString: aString .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
uint16: anInteger
	| res |
	(res := self basicNew) initializeUInt16: anInteger .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
uint32: anInteger
	| res |
	(res := self basicNew) initializeUInt32: anInteger .
	^ res
%
category: 'other'
classmethod: CPreprocessorToken
uint64: anInteger
	| res |
	(res := self basicNew) initializeUInt64: anInteger.
	^ res
%
! ------------------- Instance methods for CPreprocessorToken
category: 'other'
method: CPreprocessorToken
asAssociation
	"To facilitate automated tests"
	^type -> value.
%
category: 'other'
method: CPreprocessorToken
collectOriginalFrom: aList

	| string |
	string := String new.
	aList do: [:each | string add: each source].
	source := string.
%
category: 'other'
method: CPreprocessorToken
concatenate: aToken

	value add: (
		aToken isIdentifierToken ifTrue: [aToken value] ifFalse: [
		aToken isIntegerToken ifTrue: [aToken value printString] ifFalse: [
		self halt]]).
%
category: 'Accessing'
method: CPreprocessorToken
defineParameters

	^value key.
%
category: 'Initializing'
method: CPreprocessorToken
defineParameters: anArray string: aString

	type := #'define'.
	value := anArray -> aString.
%
category: 'Accessing'
method: CPreprocessorToken
defineString

	^value value.
%
category: 'Errors'
method: CPreprocessorToken
error: aString

  "aString is typically of the form 'Expected semicolon ...' "

  Error signal: (aString , ', found ''' , self printString , '''', self positionString)
%
category: 'Accessing'
method: CPreprocessorToken
file

	^file.
%
category: 'Reading'
method: CPreprocessorToken
hexCharacterFrom: aStream
	"Enter with pointing to 'x' of '\0xhh'"

	| string |
	string := String withAll: '16r0'.
	[true] whileTrue: [
		| char |
		char := self peekFrom: aStream.
		(string size < 6 and: [char isAlphaNumeric]) ifFalse: [
			value add: (Character codePoint: (Integer fromString: string)).
			^self.
		].
		string add: char.
		source add: aStream next.
	].
%
category: 'Initializing'
method: CPreprocessorToken
initializeArgs: anArray

	type := #'args'.
	value := anArray.
%
category: 'Initializing'
method: CPreprocessorToken
initializeConcatenation

	type := #'concatenation'.
%
category: 'Initializing'
method: CPreprocessorToken
initializeDefine: aString

	type := #'define'.
	value := nil -> aString.
%
category: 'Initializing'
method: CPreprocessorToken
initializeDirective: aString

	type := #'directive'.
	value := aString.
%
category: 'Initializing'
method: CPreprocessorToken
initializeEmpty

	type := #'empty'.
%
category: 'Initializing'
method: CPreprocessorToken
initializeIdentifier: aString

	type := #'identifier'.
	value := aString.
%
category: 'Initializing'
method: CPreprocessorToken
initializeInt16: anInteger

	type := #'int16'.
	value := anInteger.
%
category: 'Initializing'
method: CPreprocessorToken
initializeInt32: anInteger

	type := #'int32'.
	value := anInteger.
%
category: 'Initializing'
method: CPreprocessorToken
initializeInt64: anInteger

	type := #'int64'.
	value := anInteger.
%
category: 'Initializing'
method: CPreprocessorToken
initializeNumber: aNumber

	type := #'number'.
	value := aNumber.
%
category: 'Initializing'
method: CPreprocessorToken
initializePunctuator: aString

	type := #'punctuator'.
	value := aString.
%
category: 'other'
method: CPreprocessorToken
initializeStream: aStream filename: aString line: anInteger singleLogicalLine: aBoolean
		"The return value from the class-side method will be what this method returns, not always a new object"

	| isBeginningOfLine char1 char2 |
	isBeginningOfLine := self isBeginningOfLineFor: aStream.
	file := aString.
	line := anInteger.
	source := String new.
	aBoolean
		ifTrue: [self skipSingleLogicalLineCommentsInStream: aStream ifNothingFound: [^nil]]
		ifFalse: [self skipMultipleBlankLinesAndCommentsInStream: aStream].
	(char1 := self peekFrom: aStream) == nil  ifTrue: [^nil].
	char2 := aStream peek2.
	(char1 == $L and: [char2 == $']) ifTrue: [self readWideCharacterFrom: aStream] ifFalse: [
	(char1 == $L and: [char2 == $"]) ifTrue: [self readWideStringFrom: aStream] ifFalse: [
	char1 isLetter	ifTrue: [	self readIdentifierFrom: aStream] ifFalse: [
	char1 == $_	ifTrue: [	self readIdentifierFrom: aStream] ifFalse: [
	char1 == $" 	ifTrue: [	self readStringFrom: aStream] ifFalse: [
	char1 == $' 	ifTrue: [	self readCharacterFrom: aStream] ifFalse: [
	char1 == $# 	ifTrue: [	self readHashFrom: aStream isBeginningOfLine: isBeginningOfLine] ifFalse: [
	char1 isDigit 	ifTrue: [	self readNumberFrom: aStream] ifFalse: [
									self readPunctuatorFrom: aStream]]]]]]]].
	"Include trailing comments starting on this line with the source for this token"
	self skipCommentsIn: aStream.
%
category: 'Initializing'
method: CPreprocessorToken
initializeString: aString

	type := #'string'.
	value := aString.
%
category: 'Initializing'
method: CPreprocessorToken
initializeUInt16: anInteger

	type := #'uint16'.
	value := anInteger.
%
category: 'Initializing'
method: CPreprocessorToken
initializeUInt32: anInteger

	type := #'uint32'.
	value := anInteger.
%
category: 'Initializing'
method: CPreprocessorToken
initializeUInt64: anInteger

	type := #'uint64'.
	value := anInteger.
%
category: 'Testing'
method: CPreprocessorToken
isAttributeIdentifierToken

	^type == #'identifier' and: [value = '__attribute__'].
%
category: 'other'
method: CPreprocessorToken
isBeginningOfLineFor: aStream

	aStream atBeginning ifTrue: [^true].
	^aStream peek == Character lf.
%
category: 'Testing'
method: CPreprocessorToken
isCloseAngleBracketToken

	^type == #'punctuator' and: [value = '>'].
%
category: 'Testing'
method: CPreprocessorToken
isCloseCurlyBracketToken

	^type == #'punctuator' and: [value = '}'].
%
category: 'Testing'
method: CPreprocessorToken
isCloseParenthesisToken

	^type == #'punctuator' and: [value = ')'].
%
category: 'Testing'
method: CPreprocessorToken
isCloseSquareBracketToken

	^type == #'punctuator' and: [value = ']'].
%
category: 'Testing'
method: CPreprocessorToken
isColonToken

	^type == #'punctuator' and: [value = ':'].
%
category: 'Testing'
method: CPreprocessorToken
isCommaToken

	^type == #'punctuator' and: [value = ','].
%
category: 'Testing'
method: CPreprocessorToken
isConcatenationToken

	^type == #'concatenation'.
%
category: 'Testing'
method: CPreprocessorToken
isConditionalBeginDirective

	| list |
	list := #('if' 'ifdef' 'ifndef').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
method: CPreprocessorToken
isConditionalDirective

	| list |
	list := #('if' 'else' 'endif' 'ifdef' 'ifndef' 'elif').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
method: CPreprocessorToken
isConditionalEndDirective

	| list |
	list := #('else' 'endif' 'elif').
	^type == #'directive' and: [list includes: value].
%
category: 'Testing'
method: CPreprocessorToken
isDefinedIdentifierToken

	^type == #'identifier' and: [value = 'defined'].
%
category: 'Testing'
method: CPreprocessorToken
isDirectiveToken

	^type == #'directive'.
%
category: 'Testing'
method: CPreprocessorToken
isDotToken

	^type == #'punctuator' and: [value = '.'].
%
category: 'Testing'
method: CPreprocessorToken
isEmptyToken

	^type == #'empty'.
%
category: 'Testing'
method: CPreprocessorToken
isEndifDirective

	^type == #'directive' and: [value = 'endif'].
%
category: 'Testing'
method: CPreprocessorToken
isEqualsToken

	^type == #'punctuator' and: [value = '='].
%
category: 'Testing'
method: CPreprocessorToken
isExtensionIdentifierToken

	^type == #'identifier' and: [value = '__extension__'].
%
category: 'Testing'
method: CPreprocessorToken
isFunctionLikeMacroDefinition

	^type == #'define' and: [value key ~~ nil ].
%
category: 'Testing'
method: CPreprocessorToken
isHashToken

	^type == #'punctuator' and: [value = '#'].
%
category: 'Testing'
method: CPreprocessorToken
isIdentifierToken

	^type == #'identifier'.
%
category: 'Testing'
method: CPreprocessorToken
isIntegerToken

	^#(#'int16') includes: type.
%
category: 'Testing'
method: CPreprocessorToken
isOpenAngleBracketToken

	^type == #'punctuator' and: [value = '<'].
%
category: 'Testing'
method: CPreprocessorToken
isOpenCurlyBracketToken

	^type == #'punctuator' and: [value = '{'].
%
category: 'Testing'
method: CPreprocessorToken
isOpenParenthesisToken

	^type == #'punctuator' and: [value = '('].
%
category: 'Testing'
method: CPreprocessorToken
isOpenSquareBracketToken

	^type == #'punctuator' and: [value = '['].
%
category: 'Testing'
method: CPreprocessorToken
isPunctuatorToken

	^type == #'punctuator'.
%
category: 'Testing'
method: CPreprocessorToken
isReservedWord

	^#(
		'__float128'
		'_Complex'
		'auto'
		'break'
		'case'
		'char'
		'const'
		'continue'
		'default'
		'do'
		'double'
		'else'
		'entry'
		'extern'
		'float'
		'for'
		'goto'
		'if'
		'int'
		'enum'
		'long'
		'register'
		'return'
		'short'
		'signed'
		"'sizeof'"		"this should be followed by a parenthesis"
		'static'
		'struct'
		'switch'
		'typedef'
		'union'
		'unsigned'
		'void'
		'volatile'
		'while'
	) includes: value.
%
category: 'Testing'
method: CPreprocessorToken
isSemicolonToken

	^type == #'punctuator' and: [value = ';'].
%
category: 'Testing'
method: CPreprocessorToken
isStarToken

	^type == #'punctuator' and: [value = '*'].
%
category: 'Testing'
method: CPreprocessorToken
isStringToken

	^type == #'string'.
%
category: 'Testing'
method: CPreprocessorToken
isStructIdentifierToken

	^type == #'identifier' and: [value = 'struct'].
%
category: 'Testing'
method: CPreprocessorToken
isTildeToken

	^type == #'punctuator' and: [value = '~'].
%
category: 'Testing'
method: CPreprocessorToken
isUpArrowToken

	^type == #'punctuator' and: [value = '^'].
%
category: 'Testing'
method: CPreprocessorToken
isVarargsToken

	^type == #'varargs'.
%
category: 'other'
method: CPreprocessorToken
key

	^type.
%
category: 'Accessing'
method: CPreprocessorToken
line

	^line.
%
category: 'Reading'
method: CPreprocessorToken
octalCharacterFrom: aStream
	"Enter with pointing to '0' of '\0ooo'"

	| string |
	string := String withAll: '8r0'.
	[true] whileTrue: [
		| char |
		char := self peekFrom: aStream.
		(string size < 6 and: [char isDigit]) ifFalse: [
			value add: (Character codePoint: (Integer fromString: string)).
			^self.
		].
		string add: char.
		source add: aStream next.
	].
%
category: 'Accessing'
method: CPreprocessorToken
peek
	"In order to distinguish a function-like macro definition from 
	an object-like macro definition that begins with an open parenthesis,
	we need to know if there was any whitespace after the identifier.

	# define foo(bar)		/* a function-like macro */
	# define foo (bar)	/* an object-like macro */
"
	^peek
%
category: 'other'
method: CPreprocessorToken
peekFrom: aStream
	"Give back the next non-mergeLine token; does not skip whitespace, does not skip comments"

	| char1 char2 char3 lf cr |
	(char1 := aStream peek) == nil  ifTrue: [^nil].
	char1 == $\ ifFalse: [^char1].
	(char2 := aStream peek2) == nil  ifTrue: [^$\].
        lf := Character lf .
        cr := Character cr .
	(char2 == cr or: [char2 == lf]) ifFalse: [^$\].
	"$\ was followed by CR or LF"
	line := line + 1.
	source 
		add: aStream next;
		add: aStream next.
	(char3 := aStream peek) == nil  ifTrue: [^nil].
	(char2 ~~ char3 and: [char3 == cr or: [char3 == lf]]) ifFalse: [^char3].
	^self peekFrom: aStream.
%
category: 'Printing'
method: CPreprocessorToken
positionString
  | str |
  (str := ' at line ' copy) 
    add: line asString ;
    add: ' in file ' ; add: file asString .
  ^ str
%
category: 'other'
method: CPreprocessorToken
printOn: aStream

	(source == nil  or: [source isEmpty]) ifTrue: [
		type printOn: aStream.
		aStream nextPutAll: ' -> '.
		value printOn: aStream.
	] ifFalse: [
		aStream nextPutAll: source.
	].
%
category: 'Reading'
method: CPreprocessorToken
readCharacterFrom: aStream

	type := #'uint8'. 
	self 
		readLiteralUpTo: $'
		from: aStream.
	255 < (value := value first codePoint) ifTrue: [self error: 'Should be wide character!'].
%
category: 'Reading'
method: CPreprocessorToken
readHashFrom: aStream isBeginningOfLine: aBoolean

	| char |
	source add: aStream next.	"$#"
	(char := self peekFrom: aStream) == $# ifTrue: [	"## means to merge adjacent tokens"
		source add: aStream next.
		type := #'concatenation'.
		^self.
	].
	aBoolean ifFalse: [
		type := #'punctuator'.
		value := '#'.
		^self.
	].
	self 
		skipCommentsIn: aStream;
		readIdentifierFrom: aStream .
	type := #'directive'.
	value isEmpty ifTrue: [self error: 'Missing preprocessor directive'].
%
category: 'Reading'
method: CPreprocessorToken
readIdentifierFrom: aStream

	|  char |
	type := #'identifier'.
	value := String new.
	[
		char := self peekFrom: aStream.
		char ~~ nil  and: [char isAlphaNumeric or: [char == $_ or: [char == $: and: [aStream peek2 == $: or: [value last == $:]]]]].
	] whileTrue: [
		value add: aStream next.
	].
	peek := aStream peek.		"Curious to know if it is an open parenthesis"
	source add: value.
%
category: 'Reading'
method: CPreprocessorToken
readLiteralUpTo: aCharacter from: aStream

	| char lf |
	source add: aStream next.	"The open quote character"
	value := String new.
        lf := Character lf .
	[
		(char := aStream next) == lf ifTrue: [self error: 'End of literal not found on current line!'].
		source add: char.
		char = aCharacter.
	] whileFalse: [
		char == $\ ifTrue: [
			source add: (char := aStream next).
			char == $a 	ifTrue: [value add: (Character codePoint: 7	)	] ifFalse: [		"\a 	= alert, <Ctrl>+<G>"
			char == $b 	ifTrue: [value add: (Character codePoint: 8	)	] ifFalse: [		"\b 	= backspace, <Ctrl>+<H>"
			char == $f	ifTrue: [value add: (Character codePoint: 12	)	] ifFalse: [		"\f 	= form feed, <Ctrl>+<L>"
			char == $n	ifTrue: [value add: (Character codePoint: 13	)	] ifFalse: [		"\n 	= new line, <Ctrl>+<M>"
			char == $r	ifTrue: [value add: (Character codePoint: 10	)	] ifFalse: [		"\r 	= carriage return, <Ctrl>+<J>"
			char == $t	ifTrue: [value add: (Character codePoint: 9	)	] ifFalse: [		"\t 	= horizontal tab, <Ctrl>+<I>"
			char == $v	ifTrue: [value add: (Character codePoint: 11	)	] ifFalse: [		"\v 	= vertical tab, <Ctrl>+<K>"
			char == $x	ifTrue: [self hexCharacterFrom: aStream			] ifFalse: [		"\xdd = hex number"
			char == $0	ifTrue: [self octalCharacterFrom: aStream		] ifFalse: [		"\0ddd	= octal number"
			char == $'	ifTrue: [value add: char									] ifFalse: [		"\'		= single quote"
			char == $"	ifTrue: [value add: char									] ifFalse: [		"""		= double quote"
			char == $?	ifTrue: [value add: char									] ifFalse: [		"?		= question mark"
			char == $\	ifTrue: [value add: char									] ifFalse: [		"\\		= single backslash"
			]]]]]]]]]]]]].
		] ifFalse: [
			value add: char.
		].
	].
%
category: 'Reading'
method: CPreprocessorToken
readNumberFrom: aStream
"A preprocessing number has a rather bizarre definition. The category includes all the normal integer and floating point constants one expects of C, but also a number of other things one might not initially recognize as a number. Formally, preprocessing numbers begin with an optional period, a required decimal digit, and then continue with any sequence of letters, digits, underscores, periods, and exponents. Exponents are the two-character sequences `e+', `e-', `E+', `E-', `p+', `p-', `P+', and `P-'. (The exponents that begin with `p' or `P' are new to C99. They are used for hexadecimal floating-point constants.)"

	| char string isLong isUnsigned |
	type := #'number'.
	value := String with: (source add: aStream next).
	[
		aStream atEnd not and: [
			char := self peekFrom: aStream.
			char isAlphaNumeric or: [char == $. or: [char == $_ or: [(char == $+ or: [char == $-]) and: ['eEpP' includes: value last]]]].
		].
	] whileTrue: [
		value add: (source add: aStream next).
	].
	isLong := false.
	isUnsigned := false.
	(2 < value size and: [(value at: 1) == $0 and: [(value at: 2) == $x ]]) ifTrue: [
		string := '16r' , (value copyFrom: 3 to: value size).
	] ifFalse: [
		value first == $0 ifTrue: [
			string := '8r' , value.
		] ifFalse: [
			string := value.
		].
	].
	('lL' includes: string last) ifTrue: [
		isLong := true.
		string := string copyFrom: 1 to: string size - 1.
	].
	('uU' includes: string last) ifTrue: [
		isUnsigned := true.
		string := string copyFrom: 1 to: string size - 1.
		('lL' includes: string last) ifTrue: [
			isLong := true.
			string := string copyFrom: 1 to: string size - 1.
		].
	].
	value := Integer fromString: string.
	(value class ~~ SmallInteger and: [16rFFFFFFFFFFFFFFFF < value]) ifTrue: [self halt].
	type := (isLong or: [16rFFFFFFFF < value]) ifTrue: [
		(isUnsigned or: [16r7FFFFFFFFFFFFFFF < value]) 
			ifTrue: [#'uint64']
			ifFalse: [#'int64'].
	] ifFalse: [
		16rFFFF < value ifTrue: [
			(isUnsigned or: [16r7FFFFFFF < value]) 
				ifTrue: [#'uint32']
				ifFalse: [#'int32'].
		] ifFalse: [
			(isUnsigned or: [16r7FFF < value]) 
				ifTrue: [#'uint16']
				ifFalse: [#'int16'].
		].
	].
%
category: 'Reading'
method: CPreprocessorToken
readPunctuatorFrom: aStream

	| char1 char2 char3 twoChars threeChars |
	type := #'punctuator'.
	source add: (char1 := aStream next).
	value := String with: char1.
	(char2 := aStream peek) == nil  ifTrue: [^self].
	(char3 := aStream peek2) == nil  ifTrue: [
		(twoChars := value copy) add: char2 .
	] ifFalse: [
		(twoChars := value copy) add: char2 .
		(threeChars := twoChars copy) add: char3 .
	].
	(#('<<=' '>>=' '->*' '...') includes: threeChars) ifTrue: [
		value := threeChars.
		source add: char2; add: char3.
		aStream next; next.
		value = '...' ifTrue: [type := #'varargs'].
		^self.
	].
	(#('++' '--' '==' '!=' '<=' '>=' '&&' '||' '<<' '>>' '+=' '-=' '*=' '/=' '%=' '&=' '|=' '^=' '->' '.*' '::') includes: twoChars) ifTrue: [
		value := twoChars.
		source add: char2.
		aStream next.
	].
%
category: 'Reading'
method: CPreprocessorToken
readStringFrom: aStream

	type := #'string'. 
	self 
		readLiteralUpTo: $" 
		from: aStream.
%
category: 'Reading'
method: CPreprocessorToken
readWideCharacterFrom: aStream
		"#define __WCHAR_MAX__ 2147483647"

	(source add: aStream next) == $L ifFalse: [self error: 'How did we get here!?'].
	type := #'int32'.
	self 
		readLiteralUpTo: $'
		from: aStream.
	16r7FFFFFFF < (value := value first codePoint) ifTrue: [self error: 'Should be wide character!'].
%
category: 'Reading'
method: CPreprocessorToken
readWideStringFrom: aStream

	(source add: aStream next) == $L ifFalse: [self error: 'How did we get here!?'].
	type := #'string'.
	self 
		readLiteralUpTo: $' 
		from: aStream.
	self halt.
%
category: 'other'
method: CPreprocessorToken
replaceFirst: aToken

	file := aToken file.
	line := aToken line.
	source := aToken source.
%
category: 'other'
method: CPreprocessorToken
replaceOther: aToken

	file := aToken file.
	line := aToken line.
	source := ''.
%
category: 'other'
method: CPreprocessorToken
setValueToArrayWithEmptyToken

	value := { CPreprocessorToken empty }.
%
category: 'other'
method: CPreprocessorToken
skipBlockCommentsIn: aStream

	| char lf |
	source add: aStream next.	"This should be the $* following the $/"
        lf := Character lf .
	[
		source add: (char := aStream next).
		char == lf ifTrue: [line := line + 1].
		char == $* and: [(self peekFrom: aStream) == $/ ].
	] whileFalse: [].
	source add: aStream next.	"This should be the $/ following the $*"
	^self skipCommentsIn: aStream.
%
category: 'other'
method: CPreprocessorToken
skipCommentsIn: aStream
		"answer true if came to end-of-line, otherwise false"

	| char position lf |
        lf := Character lf .
	[
		(char := aStream peek) ~~ nil  and: [char isSeparator].
	] whileTrue: [
		char == lf ifTrue: [^true].
		source add: aStream next.
	].
	char == $/ ifFalse: [^false].
	position := aStream position.
	source add: aStream next.
	char := self peekFrom: aStream.
	char == $* ifTrue: [^self skipBlockCommentsIn: aStream].
	char == $/ ifTrue: [^self skipLineCommentsIn: aStream].
	source size: source size - 1.
	aStream position: position.
	^false.
%
category: 'other'
method: CPreprocessorToken
skipLineCommentsIn: aStream
        | lf |
	source add: aStream next.	"This should be the $/ following the $/"
        lf := Character lf .
	[
		aStream peek == lf.
	] whileFalse: [
		source add: aStream next.
	].
	^true.
%
category: 'other'
method: CPreprocessorToken
skipMultipleBlankLinesAndCommentsInStream: aStream

	[
		[
			self skipCommentsIn: aStream.
		] whileTrue: [
			line := line + 1.
			source add: aStream next.
		].
		aStream peek == $\.
	] whileTrue: [
		source add: aStream next.
	].
%
category: 'other'
method: CPreprocessorToken
skipSingleLogicalLineCommentsInStream: aStream ifNothingFound: aBlock
	"Skip white space and comments on the current line, extending the
	 definition of current for lines ending with a backslash+newline."
        | lf |
	[(self skipCommentsIn: aStream) ifTrue: [^aBlock value].
	(self peekFrom: aStream) == $\ and: [aStream peek2 == (lf ifNil:[lf :=Character lf]) ]]
			whileTrue: 
				[line := line + 1.
				source add: aStream next.
				source add: aStream next]
%
category: 'Accessing'
method: CPreprocessorToken
source

	source == nil  ifTrue: [^''].
	^source.
%
category: 'Accessing'
method: CPreprocessorToken
type

	^type.
%
category: 'Accessing'
method: CPreprocessorToken
value

	type == #'define' ifTrue: [self error: 'Should call #defineParameters or #defineString'].
	^value.
%
category: 'other'
method: CPreprocessorToken
value: anObject

	value := anObject.
%
category: 'other'
method: CPreprocessorToken
valueString

	type == #'empty' ifTrue: [^''].
	type == #'identifier' ifTrue: [^value].
	type == #'int16' ifTrue: [^value printString].
	type == #'int32' ifTrue: [^value printString].
	type == #'int64' ifTrue: [^value printString].
	type == #'punctuator' ifTrue: [^value].
	type == #'string' ifTrue: [^'"' , value , '"'].
	type == #'uint16' ifTrue: [^value printString].
	type == #'uint32' ifTrue: [^value printString].
	type == #'uint64' ifTrue: [^value printString].
	type == #'varargs' ifTrue: [^'...'].
	self error: 'unrecognized type: ' , type printString.
%

